/**
 * hsearch.c --- PD simple implementation of System V hsearch(3c) routine
 * It is by Arnold Robbins (arnold@skeeve.atl.ga.us) -- thanks!
**/

/**
 * Modified to deal with multiple tables 26 June 1996 -- Will Deich
 * All routines hxxx() become Hxxx().
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "Hsearch.h"

struct TABLE_CONTAINER {
    ELEMENT **Table;	/* pointer to dynamicly allocated table */
    int Num_elem;	/* number of elements */
};
typedef struct TABLE_CONTAINER TableContainer;

static int hashit();

/*
 * table of primes just below 2^n, n=2..31 for use in finding the right prime
 * number to be the table size.  this table may be generally useful...
 */

static unsigned int primetab[] = {
    /*
     * comment these out, so that table will always have a minimal size...
      3, 7, 13, 31, 61,
     */
    127, 251, 509, 1021, 2039, 4093, 8191, 16381, 32749, 65521, 131071,
    262139, 524287, 1048573, 2097143, 4194301, 8388593, 16777213, 33554393,
    67108859, 134217689, 268435399, 536870909, 1073741789, 2147483647
};

/* Hcreate --- create a hash table at least how_many big */

void *
Hcreate (int how_many)
{
    register int i, j;
    TableContainer *tab;
    void *malloc();

    /*
     * find first prime number >= how_many, and use it for table size
     */

    j = sizeof (primetab) / sizeof (primetab[0]);
    for (i = 0; i < j; i++)
	if (primetab[i] >= how_many)
	    break;

    tab = (TableContainer *) malloc(sizeof(TableContainer));
    if (!tab)
	return NULL;

    if (i >= j)	/* how_many bigger than any prime we have, use it */
	tab->Num_elem = how_many;
    else
	tab->Num_elem = primetab[i];

    if ((tab->Table = (ELEMENT **)
	    calloc ((unsigned) tab->Num_elem, sizeof (ELEMENT *))) == NULL) {
	free(tab);
	return NULL;
    } else {
	return (void *) tab;
    }
}

/* idestroy --- destroy a single element on a chain */

static void idestroy (elem)
ELEMENT *elem;
{
    if (elem != NULL)
    {
	idestroy (elem->next);
	free ((char *) elem);
    }
}

/* Hdestroy --- nuke a hash table */

void Hdestroy(void *table)
{
    register unsigned int i;
    TableContainer *tab = (TableContainer *) table;

    if (tab != NULL)
    {
	/* free all the chains */
	for (i = 0; i < tab->Num_elem; i++)
	    idestroy (tab->Table[i]);

	/* now the table itself */
	free ((char *) tab->Table);
	free (tab);
    }
}

/* Hsearch --- lookup or enter an item in the hash table */

ENTRY *Hsearch (void *table, ENTRY entry, ACTION action)
{
    ELEMENT e;
    ELEMENT *ep = NULL;
    ELEMENT *ep2 = NULL;
    TableContainer *tab = (TableContainer *) table;
    int hindex;

    if (!tab || !tab->Table)
	return (NULL);

    hindex = hashit (entry.key, tab->Num_elem);
    if (tab->Table[hindex] == NULL)	/* nothing there */
    {
	if (action == FIND)
	    return (NULL);
	else
	{
	    /* add it to the table */
	    e.item = entry;
	    e.next = NULL;
	    if ((tab->Table[hindex] = (ELEMENT *) calloc (1, sizeof (ELEMENT))) == NULL)
		return (NULL);
	    *tab->Table[hindex] = e;
	    return (& tab->Table[hindex]->item);
	}
    }
    else
    {
	/* something in bucket, see if already on chain */
	for (ep = tab->Table[hindex]; ep != NULL; ep = ep->next)
	{
	    if (strcmp (ep->item.key, entry.key) == 0)
	    {
		if (action == ENTER) 
		    ep->item.data = entry.data;
		/* already there, just change data */
		/* or action was just find it */
		return (& ep->item);
	    }
		else
		    ep2 = ep;
	}
	/* at this point, item was not in table */
	/* ep2 points at last element on the list */
	if (action == ENTER)
	{
	    if ((ep2->next = (ELEMENT *) calloc (1, sizeof (ELEMENT))) == NULL)
		return (NULL);
	    ep2->next->item = entry;
	    ep2->next->next = NULL;
	    return (& ep2->next->item);
	}
	else
	    return (NULL);
    }
    /*NOTREACHED*/
}

/* hashit --- do the hashing algorithm */

/*
* algorithm is sum of string elements, plus string length
* mod table size.
*/

static int hashit (text, num_elem)
register char *text;
int num_elem;
{
    register long int sum = 0;
    register int i;

    for (i = 0; text[i] != '\0'; i++)
	    sum += text[i];
    sum += i;

    return (sum % num_elem);
}

/*
 * Prints all elements of the table, in no particular order.
 */
void
Hprint(table, fp, f)
void *table;
FILE *fp;
void (*f)();	/* (*f) is a function that will be called by Hprint, is passed
		 * (hashIndex, ptrToKey, ptrToData), and should print these.
		 */
{
    ELEMENT *p;
    TableContainer *tab = (TableContainer *) table;
    int i;
    if (!tab)
	return;
    for (i=0; i<tab->Num_elem; i++) {
	for (p = tab->Table[i]; p; p = p->next) {
	    (*f)(fp, i, p->item.key, p->item.data);
	}
    }
}

/*
 * A function suitable for passing to Hprint, if both key and data are text
 * suitable for passing to printf.  You can use this one or supply your
 * own.
 */
void
Htext(FILE *fp, int indx, char *key, char *data)
{
    fprintf(fp, "\t%s :\t%s\n", key, data);
}

/*
 * A function suitable for passing to Hprint, if key is text and data
 * is pointer to double.  You can use this one or supply your
 * own.
 */
void
Hdouble(FILE *fp, int indx, char *key, char *data)
{
    fprintf(fp, "\t%s :\t%.12g\n", key, * (double *) data);
}

