#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>

#include "expressions.h"
#include "Hsearch.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef M_E
#define M_E 2.7182818284590452354
#endif

#define FMT "\t%.12g\n"

#define deg2rad(x) ((x) * M_PI / 180.)
#define rad2deg(x) ((x) * 180. / M_PI)

#define NSTACK 100

/* ===================================================================
 * NOTE: in the following, X means top of stack, Y means next on stack
 * ===================================================================
 */

/* ------------------------------------------------------------------- */
/* A struct OPS is used to describe each builtin operation. */
struct OPS {
    char *name;		/* Name of operator -- this is what the user must
			 * enter to invoke the operator.  If "", then this
			 * is simply a placeholder for some help info
			 * (see "info" member, below).
			 */
    double (*f)();
    int arglist;	/* Code for argument list:
			 * Nil -- no arguments;
			 * Stack -- pass (ptr to top of stack, last elt);
			 * Top -- copy top arg, pass to func.
			 * Pop1 -- pop 1 arg, pass to func.
			 * Pop2 -- pop 2 args, pass (y, x) to func.
			 */
    int result;		/* Code for what to do with result:
			 *	Nil -- ignore result;
			 *	Clear -- clear stack;
			 *	Push -- push onto stack;
			 *	Exch -- exchange top two elements on stack;
			 */
    char *info;		/* one line of help text -- displayed when user
			 * uses the "long help" help command.
			 */

};
typedef struct OPS Ops;

/* Tokens look like  str   or  str,arg */
struct token {
    char *s;		/* ptr to start of str */
    int len;		/* length of str (not including ",arg") */
    char *arg;		/* ptr to start of arg */
    int arglen;		/* length of arg */
};
typedef struct token Token;

/* ------------------------------------------------------------------- */

static char *token_s;		/* where the tokenizer saves its pointer to
				 * the next-position-to-read in input buffer.
				 * If you want nested calls, save this
				 * and restore it around the nested call.
				 */
char * (*rpn_fgets)(char *buf, int len, FILE *fp) = &fgets;
				/* Input function.  Normally, fgets() is used
				 * for input.  This lets caller supply alt.
				 */

static double stack[NSTACK];	/* bottom of stack is stack[1]; top of stack
				 * is stack[i], i>=0.  Note that top=&stack[0]
				 * means empty stack, and that there are thus
				 * NSTACK-1 available elements since the 0'th
				 * element marks an empty stack.
				 */

static double lastx;		/* Stores the last 'X' (top of stack) value
				 * before current number or operator was
				 * processed.
				 */
static void *macros=NULL;	/* user-defined macros. */
static void *variables=NULL;	/* user-defined variables. */

static char printfmt[100] = FMT;	/* for printing stack elements */
static int fmt_is_int = 0;		/* treat data as integer-like? */

static int debug=0;		/* debug level */

static int trig_mode_is_rad=1;	/* !0 -> degrees */

Ops *thisop = NULL;		/* Current operator */

Token *thistok = NULL;		/* hack: ptr to most-recently returned token,
				 * so that some operators can extract
				 * the arg field (recall that tokens look like
				 * str or str,arg).
				 */

static Ops *findop(Token tok);	/* Returns ptr to operator */

/* ------------------------------------------------------------------- */
/* See the description of the following codes under the declaration for
 * struct OPS, above.
 */
enum Code { Nil = -99, Top, Stack, Pop1, Pop2, Push, Exch, Clear };

/* ------------------------------------------------------------------- */

static double Add(double y, double x) { return y+x; }
static double Sub(double y, double x) { return y-x; }
static double Mul(double y, double x) { return y*x; }
static double Div(double y, double x) { return y/x; }
static double Bitor(double y, double x) {
		return (double) ((long) x | (long) y);}
static double Bitxor(double y, double x) {
		return (double) ((long) x ^ (long) y);}
static double Bitand(double y, double x) {
		return (double) ((long) x & (long) y);}
static double Compl(double x) { return (double) (~ (long) x); }
static double Shiftl(double y, double x) {
		return (double) ((long) y << (long) x);}
static double Shiftr(double y, double x) {
		return (double) ((long) y >> (long) x);}
static double Aint(double x) { return x - fmod(x, 1.) ; }
static double Nint(double x) { return floor(x+0.5) ; }
static double Nfloat(double x) { float f = (float) x; return ((double) f); }
static double Lg(double x) { return log(x)/log(2.); }
static double Sqr(double x) { return x * x; }
static double Asinh(double x) { return log(x + sqrt(x*x + 1)); }
static double Acosh(double x) { return (x<1.) ? 0. : log(x + sqrt(x*x - 1)); }
static double Atanh(double x) { return ((x*x)>=1.) ? 0. : 0.5*log((1.+x)/(1.-x)); }
static double Exp10(double x) { return pow(10., x); }
static double Chs(double x) { return -x; }
static double Pi(void) { return M_PI; }
static double E(void) { return M_E; }
static double nil(void) { return 0.; }
static double Invt(double x) { return 1./x; }
static double Frac(double x) { return fmod(x, 1.); }
static double Dup(double x) { return x; }
static double Hms(double x) { double h=Aint(x); double m=Aint((x-h)*60.);
		double s=(x-h-(m/60.))*3600.; return h*10000. + m*100. + s ; }
static double Hours(double x) {
		double s=fmod(x,100.);
		double m=fmod((x-s)/100, 100);
		return Aint(x/10000.) + m/60. + s/3600.; }
static double Degrad(double x) { return deg2rad(x) ; }
static double Raddeg(double x) { return rad2deg(x) ; }
static double Debug(double);
static double Dec(void)   { strcpy(printfmt, "\t%d\n");   fmt_is_int=1; }
static double Octal(void) { strcpy(printfmt, "\t%#lo\n"); fmt_is_int=1; }
static double Hex(void)   { strcpy(printfmt, "\t%#lx\n"); fmt_is_int=1; }
static double Sci(double x) { int pre = (int) (x+0.5);
		if (x < 0) strcpy(printfmt, "\t%g\n");
		else       sprintf(printfmt, "\t%%.%dg\n", pre);
		fmt_is_int=0; }
static double Fix(double x) { int pre = (int) (x+0.5);
		if (x < 0) strcpy(printfmt, "\t%f\n");
		else       sprintf(printfmt, "\t%%.%df\n", pre);
		fmt_is_int=0; }
static double Prttop(double *top, double *bot) { rpn_print(NULL,*top);
		return *top; }
static double Prtstk(double *top, double *bot) {
		for (top++; top-- > bot; ) rpn_print(NULL, *top);
		fputs("\n", stdout); return *top; }
static double Rollr(double *top, double *bot) { double temp=*top;
		for (;top > bot; top--) *top = *(top-1); *bot = temp; }
static double Lastx(void) {return lastx; }
static double ListH(void);
static double ListHLong(void);
static double Listmacros(void);
static double Listvariables(void);
static double Debug(double level) { debug = (int) (level + 0.5); }
static double Rad(void) { trig_mode_is_rad = 1; }
static double Deg(void) { trig_mode_is_rad = 0; }
static double Sin(double x) {return sin(trig_mode_is_rad? x : deg2rad(x)); }
static double Cos(double x) {return cos(trig_mode_is_rad? x : deg2rad(x)); }
static double Tan(double x) {return tan(trig_mode_is_rad? x : deg2rad(x)); }
static double Asin(double x) {double d=asin(x);
		return ((trig_mode_is_rad) ?  d : rad2deg(d)); }
static double Acos(double x) {double d=acos(x);
		return ((trig_mode_is_rad) ?  d : rad2deg(d)); }
static double Atan(double x) {double d=atan(x);
		return ((trig_mode_is_rad) ?  d : rad2deg(d)); }
/* ------------------------------------------------------------------- */

static Ops ops[] = {
    { "",	0,	0,	0,    "Math:        " },
    { "+",	Add,	Pop2,	Push, "Replaces top two elts with Y+X" },
    { "-",	Sub,	Pop2,	Push, "Replaces top two elts with Y-X" },
    { "*",	Mul,	Pop2,	Push, "Replaces top two elts with Y*X" },
    { "x",	Mul,	Pop2,	Push, "Alias for *" },
    { "/",	Div,	Pop2,	Push, "Replaces top two elts with Y/X" },
    { "**",	pow,	Pop2,	Push, "Replaces top two elts with pow(Y,X)" },
    { "pow",	pow,	Pop2,	Push, "Alias for '**'\n" },
    { "%",	fmod,	Pop2,	Push, "Replaces top two elts with mod(Y,X)" },
    { "mod",	fmod,	Pop2,	Push, "Alias for %" },
    { "fmod",	fmod,	Pop2,	Push, "alias for %" },

    { "",	0,	0,	0,    "Binary Bitwise:" },
    { "|",	Bitor,	Pop2,	Push, "Bitwise OR of top 2 elements "
					"(converted to long)" },
    { "&",	Bitand,	Pop2,	Push, "Bitwise AND of top 2 elements "
					"(converted to long)" },
    { "^",	Bitxor,	Pop2,	Push, "Bitwise XOR of top 2 elements "
					"(converted to long)" },
    { ">>",	Shiftr,	Pop2,	Push, "Converts top two elts to long, "
					"replaces them with Y>>X" },
    { "<<",	Shiftl,	Pop2,	Push, "Converts top two elts to long, "
					"replaces them with Y<<X" },
    { "~",	Compl,	Pop1,	Push, "Converts top elt to long, replaces "
					"it with bitwise complement\n"},

    { "",	0,	0,	0,    "Basic Unary:" },
    { "chs",	Chs,	Pop1,	Push, "Changes the sign of the top element" },
    { "abs",	fabs,	Pop1,	Push, "Replace the top elt with |X|" },
    { "inv",	Invt,	Pop1,	Push, "Replace the top elt with 1/X" },
    { "invt",	Invt,	Pop1,	Push, "Alias for inv" },
    { "frac",	Frac,	Pop1,	Push, "Replace the top elt with its "
					"fractional part" },
    { "aint",	Aint,	Pop1,	Push, "Replace the top elt with its "
					"integral part" },
    { "int",	Aint,	Pop1,	Push, "Alias for aint" },
    { "nint",	Nint,	Pop1,	Push, "Replace the top elt with the "
					"nearest integer" },
    { "nfloat",	Nfloat,	Pop1,	Push, "Replace the top elt with the "
					"nearest float value" },
    { "ceil",	ceil,	Pop1,	Push, "Replace the top elt with the "
					"smallest integer >= X" },
    { "floor",	floor,	Pop1,	Push, "Replace the top elt with the "
					"largest integer <= X\n" },
    { "sqrt",	sqrt,	Pop1,	Push, "Replaces top elt with its square root" },
    { "sqr",	Sqr,	Pop1,	Push, "Replaces top elt with X*X\n" },


    { "",	0,	0,	0,    "Trigonometric:" },
    { "sin",	Sin,	Pop1,	Push, "Replaces top elt with sin(X)" },
    { "cos",	Cos,	Pop1,	Push, "Replaces top elt with cos(X)" },
    { "tan",	Tan,	Pop1,	Push, "Replaces top elt with tan(X)" },
    { "asin",	Asin,	Pop1,	Push, "Replaces top elt with asin(X)" },
    { "acos",	Acos,	Pop1,	Push, "Replaces top elt with acos(X)" },
    { "atan",	Atan,	Pop1,	Push, "Replaces top elt with atan(X)" },
    { "atan2",	atan2,	Pop2,	Push, "Replaces top two elts with "
					"atan2(Y,X)\n" },

    { "",	0,	0,	0,    "Hyperbolic:" },
    { "sinh",	sinh,	Pop1,	Push, "Replaces top elt with sinh(X)" },
    { "cosh",	cosh,	Pop1,	Push, "Replaces top elt with cosh(X)" },
    { "tanh",	tanh,	Pop1,	Push, "Replaces top elt with tanh(X)" },
    { "asinh",	Asinh,	Pop1,	Push, "Replaces top elt with asinh(X)" },
    { "acosh",	Acosh,	Pop1,	Push, "Replaces top elt with acosh(X)" },
    { "atanh",	Atanh,	Pop1,	Push, "Replaces top elt with atanh(X)\n" },

    { "",	0,	0,	0,    "Exp. and Log.:" },
    { "exp",	exp,	Pop1,	Push, "Replaces top elt with exp(X)" },
    { "exp10",	Exp10,	Pop1,	Push, "Replaces top elt with 10^X" },
    { "ln",	log,	Pop1,	Push, "Replaces top elt with ln(X)" },
    { "log",	log,	Pop1,	Push, "Alias for ln" },
    { "lg",	Lg,	Pop1,	Push, "Replaces top elt with log_2(X)" },
    { "log10",	log10,	Pop1,	Push, "Replaces top elt with log_10(X)\n" },

    { "",	0,	0,	0,    "Conversions:" },
    { "degrad",	Degrad,	Pop1,	Push, "Convert top element (assumed deg) to "
					"radians" },
    { "raddeg",	Raddeg,	Pop1,	Push, "Convert top element (assumed rad) to "
					"degrees\n" },
    { "hms",	Hms,	Pop1,	Push, "Top elt is decimal hrs; convert to "
					"hhmmss.sss" },
    { "hours",	Hours,	Pop1,	Push, "Top elt is hhmmss.sss; convert to "
					"decimal hrs" },
    { "hrs",	Hours,	Pop1,	Push, "Alias for hours\n" },

    { "",	0,	0,	0,    "Stack Ops:" },
    { "exch",	nil,	Nil,	Exch, "Exchange the top two elements of "
					"the stack" },
    { "xfy",	nil,	Nil,	Exch, "Alias for exch" },
    { "pop",	nil,	Pop1,	Nil,  "Pop the stack" },
    { "dup",	Dup,	Top,	Push, "Duplicate the top elt on the stack" },
    { "r",	Rollr,	Stack,	Nil,  "Rolls stack by one elt" },
    { "lastx",	Lastx,	Nil,	Push, "Pushes 'last-X' value onto the stack" },
    { "clear",	nil,	Nil,	Clear,"Delete all elements from the stack\n" },

    { "",	0,	0,	0,    "Constants:" },
    { "pi",	Pi,	Nil,	Push, "Push Pi onto the stack" },
    { "e",	E,	Nil,	Push, "Push e onto the stack\n" },

    { "",	0,	0,	0,    "Display, Units:" },
    { "rad",	Rad,	Pop1,	Push, "Trig functions use/return radians" },
    { "deg",	Deg,	Pop1,	Push, "Trig functions use/return degrees" },
    { "sci",	Sci,	Pop1,	Nil,  "Print in sci format with precision=X; "
					"pop stack" },
    { "fix",	Fix,	Pop1,	Nil,  "Print in fixed format with precision=X; "
					"pop stack" },
    { "dec",	Dec,	Nil,	Nil,  "Input integral values, "
					"print decimal format" },
    { "decimal",Dec,	Pop1,	Nil,  "Alias for dec" },
    { "octal",	Octal,	Nil,	Nil,  "Input integral values, "
					"print octal format" },
    { "hex",	Hex,	Nil,	Nil,  "Input integral values, "
					"print hexadecimal format" },
    { "hexadecimal",	Hex,	Nil,	Nil, "Alias for hex\n" },

    { "",	0,	0,	0,    "Other:\t" },
    { "p",	Prttop,	Stack,	Nil,  "Print the top element on the stack" },
    { "=",	Prtstk,	Stack,	Nil,  "Print the entire stack" },
    { "h",	ListH,Nil,	Nil,  "Short help: list all operations" },
    { "?",	ListH,Nil,	Nil,  "Alias for h" },
    { "help",	ListH,Nil,	Nil,  "Alias for h" },
    { "H",	ListHLong,Nil,	Nil,  "`Long' help (one line per operator)" },
    { "Help",	ListHLong,Nil,	Nil,  "Alias for H" },
    { "macros",	Listmacros,Nil,	Nil,  "List all macros" },
    { "M",	Listmacros,Nil,	Nil,  "Alias for `macros'" },
    { "variables",	Listvariables,Nil,	Nil, "List all variables" },
    { "vars",	Listvariables,Nil,	Nil, "Alias for `variables'" },
    { "V",	Listvariables,Nil,	Nil, "Alias for `variables'" },
    { "debug",	Debug,	Pop1,	Nil,  "Set debug level to X; pop stack" },
    { NULL,	NULL,	Nil,	Nil },
};
/* ------------------------------------------------------------------- */

/*
 * Operator to implement short help operator.
 */
static double ListH(void) {
    Ops *op, *prev;
    int nout;

    if (thistok->arg) {
	/* Help on an argument */
	Token tok;
	if (!thistok->arglen) {
	    printf("%s: empty argument.  Use  %s,<HelpItem>, e.g. %s,sqrt\n",
			    thisop->name, thisop->name, thisop->name);
	    return 0.;
	}
	/* Build token to find, then search for it */
	tok.s = thistok->arg;
	tok.len = thistok->arglen;
	tok.arg = NULL;
	tok.arglen = 0;
	op = findop(tok);

	if (op) {
	    printf("    %s\t%s\n", op->name, op->info);
	} else {
	    printf("    Unknown help item <%.*s>\n", tok.len, tok.s);
	}
    } else {
	/* No argument given, so list all */
	for (op=ops, nout=0, prev=NULL; op->name; prev = op, op++) {
	    if (*op->name == '\0') {
		printf("\n%s\t", op->info);
		nout = 0;
	    } else {
		nout++;
		if (prev->f == op->f &&
			    prev->arglist == op->arglist &&
				    prev->result == op->result) {
		    fputc('|', stdout);
		} else {
		    fputs("  ", stdout);
		    if (nout >= 9 && (op+1)->name && *(op+1)->name != '\0') {
			fputs("\n\t\t\t", stdout);
			nout = 1;
		    }
		}
		fputs(op->name, stdout);
	    }
	}
	fputc('\n', stdout);
	printf("Macros:\t\t  Dxxx ...txt... (Define macro xxx)   "
				    "Exxx (Expand macro)\n");
	printf("Variables:\t  Sxxx (Copy top of stack to xxx)     "
				    "$xxx (Push xxx onto stack)\n");
	printf("Files:\t\t  @file (Execute contents of file)\n");
	printf(       "\t\t  @@file (@file plus echo lines)\n");
	printf("Numbers:\t  xxx (Push number xxx onto stack)\n");
	printf("Comments:\t  #...\tRest of line is comment\n");
	printf("\n");
    }
    return 0.;
}
/* ------------------------------------------------------------------- */

/*
 * Operator to implement long help operator.
 */
static double ListHLong(void) {
    Ops *op;
    for (op=ops; op->name; op++)
	if (*op->name == '\0')
	    puts(op->info);
	else
	    printf("\t%s\t%s\n", op->name, op->info);
    printf("\n");
    printf("Dxxx ...\tDefine macro xxx to be rest of string\n");
    printf("Exxx\tExpand macro xxx\n");
    printf("Sxxx\t(Store xxx) Copy top of stack to variable xxx\n");
    printf("$xxx\tPush variable xxx onto stack\n");
    printf("xxx\tPush number xxx onto stack\n");
    printf("#...\tComment\n");
    printf("\n");
    return 0.;
}
/* ------------------------------------------------------------------- */

static void
rpn_print_macro(FILE *fp, int indx, char *key, char *data)
{ fprintf(fp, "  %s : %s\n", key+1, data); }

/*
 * Operator to list macros.
 */
static double Listmacros(void) {
    Hprint(macros, stdout, rpn_print_macro);
    return 0.;
}
/* ------------------------------------------------------------------- */

static void
rpn_print_variable(FILE *fp, int indx, char *key, char *data)
{ fprintf(fp, "  %s : %.12g\n", key+1, *(double *) data); }

/*
 * Operator to list variables.
 */
static double Listvariables(void) {
    Hprint(variables, stdout, rpn_print_variable);
    return 0.;
}
/* ------------------------------------------------------------------- */

/*
 * Function to look up an operator function, given a string name.
 */
static Ops *
findop(Token tok)
{
    Ops *op;
    char *p;
    
    for (op = ops; op->name && (strncmp(tok.s, op->name, tok.len) != 0 ||
					    op->name[tok.len] != '\0') ; op++)
	;
    if (op->name) {
	thisop = op;
	return op;
    } else {
	thisop = NULL;
	return NULL;
    }
}
/* ------------------------------------------------------------------- */

/*
 * Returns pointer to hash table of macro definitions.
 * Can be used by an outside routine.
 */
void *
rpn_macros()
{
    return macros;
}
/* ------------------------------------------------------------------- */

/*
 * Returns pointer to stack.  See top of file for description for
 * stack layout.
 * Can be used by an outside routine.
 */
double *
rpn_stack()
{
    return stack;
}
/* ------------------------------------------------------------------- */

/*
 * Returns pointer to variables table (null if none yet defined).
 * Can be used by an outside routine.
 */
void *
rpn_variables()
{
    return variables;
}
/* ------------------------------------------------------------------- */

/*
 * Function to print a value, either into a string or to stdout.
 * Can be used by an outside routine.
 */
char *
rpn_print(char *buf, double x)
{
    if (buf) {
	if (fmt_is_int)
	    sprintf(buf, printfmt, (long) x);
	else
	    sprintf(buf, printfmt, x);
    } else {
	if (fmt_is_int)  
	    printf(printfmt, (long) x);
	else
	    printf(printfmt, x);
    }
    return buf;
}

/* ------------------------------------------------------------------- */
/*
 * Function to define a macro.
 */
static void
defmacro(Token tok,	/* Token string is Dname ... ;everything after the Dname
			 * is the definition.
			 */
	char *prefix,
	char *msg)
{
    ENTRY e;
    int k, l;

    if (debug) printf("Entering defmacro()\n");

    *msg = '\0';

    if (tok.len < 2) {
	sprintf(msg, "%s: bad macro defn -- need name after 'D'", prefix);
	    if (debug) printf("Returning from defmacro()\n");
	return;
    }

    if (!macros && !(macros = Hcreate(50))) {
	sprintf(msg, "%s: failed to allocate space for macros table",
		prefix);
	if (debug) printf("Returning from defmacro()\n");
	return;
    }

    if (!(e.key = malloc(strlen(tok.s)+1))) {
	sprintf(msg, "%s: failed to allocate space for macro entry",
		prefix);
	if (debug) printf("Returning from defmacro()\n");
	return;
    }
    strcpy(e.key, tok.s);
    /* Null-terminate the macro name.  Note that the tok.len includes the
     * leading 'D'.
     */
    e.key[tok.len] = '\0';
    /* The data immediately follows the macro name */
    e.data = e.key + tok.len + 1;

    if (!Hsearch(macros, e, ENTER)) {
	sprintf(msg, "%s: failed to insert macro in table",
		prefix);
	if (debug) printf("Returning from defmacro()\n");
	return;
    }

    if (debug) printf("Returning from defmacro()\n");
}

/* ------------------------------------------------------------------- */
/*
 * Function to define a variable.
 */
static void
defvariable(Token tok,	/* Token string is Sname */
	double val,
	char *prefix,
	char *msg)
{
    ENTRY e;
    int k, l;

    if (debug) printf("Entering defvariable()\n");

    *msg = '\0';

    if (tok.len < 2) {
	sprintf(msg, "%s: bad variable defn -- need name after 'S'", prefix);
	    if (debug) printf("Returning from defvariable()\n");
	return;
    }

    if (!variables && !(variables = Hcreate(50))) {
	sprintf(msg, "%s: failed to allocate space for variables table",
		prefix);
	if (debug) printf("Returning from defvariable()\n");
	return;


    } else if (!(e.key = malloc(tok.len+1)) ||
				!(e.data = malloc(sizeof(double)))) {
	sprintf(msg, "%s: failed to allocate space for variable entry",
		prefix);
	if (debug) printf("Returning from defvariable()\n");
	return;
    }
    strncpy(e.key, tok.s, tok.len);
    /* Null-terminate the variable name.  Note that the tok.len includes the
     * leading 'S'.
     */
    e.key[tok.len] = '\0';

    memcpy((void *) e.data, (void *) &val, sizeof(double));

    if (!Hsearch(variables, e, ENTER)) {
	sprintf(msg, "%s: failed to insert variable in table",
		prefix);
	if (debug) printf("Returning from defvariable()\n");
	return;
    }

    if (debug) printf("Returning from defvariable()\n");
}

/* ------------------------------------------------------------------- */
/*
 * Function to expand a macro invocation and evaluate it (macros are expanded
 * using Ex, where x is any character except '\0').
 */
static void
expandmacro(Token tok,	/* Expandstring is Ex */
	int flags,	/* RPN_INIT is discarded from these */
	char *prefix,
	char *msg)
{
    *msg = '\0';
    if (debug) printf("Entering expandmacro()\n");
    if (tok.len < 2) {
	sprintf(msg, "%s: bad macro use -- need a name after 'E'", prefix);
    } else {
	ENTRY e;
	ENTRY *ep;
	char save_char[2];

	e.key = tok.s;
	e.data = NULL;
	save_char[0] = tok.s[0];
	tok.s[0] = 'D';
	save_char[1] = tok.s[tok.len];
	tok.s[tok.len] = '\0';
	
	ep = Hsearch(macros, e, FIND);
	tok.s[0] = save_char[0];
	tok.s[tok.len] = save_char[1];

	if (!ep) {
	    (void) sprintf(msg, "%s: macro `%.*s' is not defined",
					    prefix, tok.len-1, tok.s+1);
	    return;
	}

	/* macro is defined -- evaluate it */ 
	flags &= ~RPN_INIT;
	(void) rpn_expr(0., flags, &ep->data, 1, prefix, msg);
    }
    if (debug) printf("Returning from expandmacro()\n");
}
/* ------------------------------------------------------------------- */

/*
 * Wrapper for function that evaluates a string containing an rpn expression.
 * This function can be used by an outside routine.
 */
double
rpn_expr(double val,	/* current value: iff the RPN_INIT flag is set,
			 * val is pushed onto the stack before evaluating
			 * the expression.
			 */
	int flags,	/* Or'd combination of (see expressions.h):
			 *	RPN_INIT
			 *	RPN_AUTOPRINT
			 * We do NOT pay attention to the RPN_ECHO
			 * flag; we assume that flag is for caller routines
			 * processing lines at a time.
			 */
	char *exprs[],	/* RPN expression to evaluate -- total expression
			 * is equiv to elements concatenated with spaces
			 * between. */
	int nexprs,	/* number of elements in exprs[] */
	char *prefix,	/* if err, write prefix: before rest of message */
	char *msg)	/* on error, write message here; '\0' on success.
			 * msg is NOT newline-terminated.
			 */
{
    char *saved_s;
    double d;
    static double rpn_expr_body(double val, int flags,
		    char *exprs[], int nexprs,
		    char *prefix, char *msg, int depth, int *modified_stack);
    static int modified_stack=0;
    static int depth=0;

    *msg = '\0';

    if (depth==0)
	modified_stack = 0;

    saved_s = token_s;
    depth++;
    d = rpn_expr_body(val, flags, exprs, nexprs,
			prefix, msg, depth, &modified_stack);
    depth--;
    token_s = saved_s;


    return d;
}

/*
 * Function to evaluate a string containing an rpn expression.
 * This function has a lot of syntax knowledge built into it.
 * Must NOT be used by an outside routine: it needs a wrapper
 * (you should use rpn_expr()) to handle state info.
 * See rpn_expr() for the meaning of the arguments.
 * The extra pair of arguments indicates (a) the call nesting depth
 * and (b) if the stack has been modified since
 * the last print command or since entering rpn_expr_body().
 */
static double
rpn_expr_body(double val, int flags, char *exprs[],
			int nexprs, char *prefix, char *msg,
			int depth, int *modified_stack)
{
    char c;
    int iexpr, len, n;
    unsigned int u;
    double x, y;
    Ops *op;
    char *p, *expr;
    Token tok;
    static Token tokenize(char *str, double *d);
    int is_defn;
    int pushed_a_value=0;
    static double *top = &stack[0];	/* Note that stack[0] is the
					 * nothing-on-stack marker.
					 */

    #define CLEAR { top = &stack[0]; *top = 0.; }

    /* Don't use top in the argument to PUSH!! */
    #define PUSH(d) {							\
	if (top == &stack[NSTACK-1]) {					\
	    sprintf(msg, "%s: stack overflow -- clearing stack!", prefix);\
	    CLEAR;							\
	    if (debug) printf("Returning from rpn_expr_body()\n");	\
	    return 0.;							\
	}								\
	*++top = d;							\
    }

    /* Don't use top in the argument to POP!! */
    /* Be silent on underflow... */
    #define POP(x) {							\
	x = *top;							\
	if (top > &stack[0])						\
	    top--;							\
	else								\
	    *top = 0.;	/* Clear top element if we pop it */		\
    }

    if (debug) printf("Entering rpn_expr_body()\n");

    /* Initialize... */
    if (flags & RPN_INIT) {
	CLEAR;
	PUSH(val);
    }

    for (iexpr=0; iexpr < nexprs;  iexpr++) {
	for (op=NULL, tok = tokenize(exprs[iexpr], &x);
				    tok.s; tok = tokenize(NULL, &x)) {
	    double new_lastx = *top;

	    /* Save last X value -- don't overwrite lastx value yet, because
	     * we need it if the lastx operator is invoked.
	     */
	    thistok = &tok;
	    thisop = NULL;

	    if (debug)
		printf("tok.s = <%.*s>(%d) arg=<%.*s>(%d)\n",
			    tok.len<0 ? -tok.len : tok.len, tok.s, tok.len,
			    tok.arglen, tok.arg, tok.arglen);
	    if (tok.len == 0) {
		/* Invalid number */
		sprintf(msg, "%s: Invalid %s number at: %.10s...",
			prefix, fmt_is_int ? "integer" : "float", tok.s);
		if (debug) printf("Returning from rpn_expr_body()\n");
		return 0.;
	    } else if (tok.len < 0) {
		/* Valid number */
		if (debug) printf("Push number (%f)\n", x);
		*modified_stack = 1;
		PUSH(x);
	    } else if (*tok.s == '#') {
		/* Rest of line is comment */
		if (debug) printf("Reached comment (%s)\n", tok.s);
		if (debug) printf("Returning from rpn_expr_body()\n");
		return *top;
	    } else if (*tok.s == 'D') {
		if (debug) printf("Define macro %.*s\n", tok.len, tok.s);
		defmacro(tok, prefix, msg);
		if (*msg) {
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}
		/* skip rest of line -- it's a definition */
		return *top;
	    } else if (*tok.s == 'E') {
		if (debug) printf("Expand macro %.*s\n", tok.len, tok.s);
		expandmacro(tok, flags, prefix, msg);
		if (*msg) {
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}
	    } else if (*tok.s == '@') {
		/* It's @filename or @@filename */
		if (debug) printf("Load file: %.*s\n", tok.len, tok.s);
		if (tok.s[1] == '@' && tok.len <= 2) {
		    sprintf(msg, "%s: @@xxx is missing filename xxx", prefix);
		    return 0.;
		} else if (tok.len <= 1) {
		    sprintf(msg, "%s: @xxx is missing filename xxx", prefix);
		    return 0.;
		}
		/* Null-terminate filename, call rpn_file, then replace
		 * overwritten character.
		 */
		{
		    char saved_c = tok.s[tok.len];
		    int echo = (tok.s[1] == '@');
		    char *thisfile = (echo) ? tok.s+2 : tok.s+1;
		    int ourflags;
		    tok.s[tok.len] = '\0';

		    ourflags = flags;
		    if (echo)
			ourflags |= RPN_ECHO;

		    rpn_file(0., ourflags, thisfile, NULL, 0, prefix, msg);
		    tok.s[tok.len] = saved_c;
		}
		if (*msg) {
		    puts(msg);
		    if (errno)
			perror("");
		    *msg = '\0';
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}
	    } else if (*tok.s == 'S') {
		if (debug) printf("Store variable %.*s\n", tok.len, tok.s);
		if (tok.len < 2) {
		    sprintf(msg, "%s: Sxxx is missing name xxx", prefix);
		    return 0.;
		}
		defvariable(tok, (top) ? *top : 0.0, prefix, msg);
		if (*msg) {
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}

	    } else if (*tok.s == '$') {

		ENTRY e;
		ENTRY *ep;
		char save_char[2];

		if (debug) printf("Push variable %.*s\n", tok.len, tok.s);

		e.key = tok.s;
		e.data = NULL;
		save_char[0] = tok.s[0];
		tok.s[0] = 'S';
		save_char[1] = tok.s[tok.len];
		tok.s[tok.len] = '\0';
		
		ep = Hsearch(variables, e, FIND);
		tok.s[0] = save_char[0];
		tok.s[tok.len] = save_char[1];

		if (!ep) {
		    (void) sprintf(msg, "%s: variable `%.*s' is not defined",
					    prefix, tok.len-1, tok.s+1);
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}

		x = *((double *) ep->data);
		PUSH(x);
		*modified_stack = 1;

	    } else if ((op = findop(tok)) != NULL) {
		/* It's an op */

		if (debug) printf("Matched op name=<%s>\n", op->name);
		switch(op->arglist) {
		case Nil:
		    x = op->f(); break;
		case Pop1:
		    POP(x) ; x = op->f(x); break;
		case Pop2:
		    POP(x); POP(y) ; x = op->f(y, x); break;
		case Top:
		    x = *top ; x = op->f(x); break;
		case Stack:
		    x = op->f(top, &stack[1]); break;
		default:
		    sprintf(msg, "%s: Internal error: unknown arglist code %d",
				    prefix, op->arglist);
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}
		switch(op->result) {
		case Nil:
		    break;
		case Push:
		    PUSH(x); break;
		case Exch:
		    POP(x); POP(y) ; PUSH(x) ; PUSH(y) ; break;
		case Clear:
		    CLEAR; *top = 0.; break;
		default:
		    sprintf(msg, "%s: Internal error: unknown result code %d",
				    prefix, op->arglist);
		    if (debug) printf("Returning from rpn_expr_body()\n");
		    return 0.;
		}
		if (op->f == Prtstk || op->f == Prttop) {
		    /* last action was to print */
		    *modified_stack = 0;
		} else if (op->arglist == Nil && op->result == Nil) {
		    /* Nil args and Nil operations on the result are
		     * in general unlikely to have modified the stack.
		     */
		    ;
		} else {
		    /* Other operations will have modified the stack */
		    *modified_stack = 1;
		}

	    } else {
		if (tok.len && *tok.s == ',')
		    sprintf(msg, "%s: Tokens may not begin with comma: %.10s",
					prefix, tok.s);
		else if (tok.arglen)
		    sprintf(msg, "%s: unknown token <%.*s> (arg=<%.*s>)",
					prefix, tok.len, tok.s,
					tok.arglen, tok.arg);
		else
		    sprintf(msg, "%s: unknown token <%.*s>",
					prefix, tok.len, tok.s);
		if (debug) printf("Returning from rpn_expr_body()\n");
		return 0.;
	    }
	    /* Do debugging if last cmd wasn't stack print. */
	    if (debug && !(op && op->f == Prtstk)) {
		int n = ((top-debug) > &stack[1]) ? debug-1 : top - &stack[1];
		printf("Top of stack (%d values on stack; max allowed=%d)\n",
				top+1-&stack[1], NSTACK-1);
		Prtstk(top, top - n);
	    }
	    lastx = new_lastx;
	}
    }
    *msg = '\0';

    if ((depth == 1) && (flags & RPN_AUTOPRINT) && *modified_stack) {
	if (fmt_is_int)
	    printf(printfmt, (long) *top);
	else
	    printf(printfmt, *top);
    }
    if (debug) printf("Returning from rpn_expr_body()\n");
    return *top;
}

/* Finds next token.  For numbers: if it's an invalid number, the length field
 * is returned 0; if it's a valid number, then length field is returned
 * with -n, where |n| is the length of the number.
 */
static Token
tokenize(char *str,	/* input string, or NULL to continue */
	double *d)	/* if number, returned with value */
{
    char *p;
    Token tok;

    /* Save new strings */
    if (str)
	token_s = str;

    /* Skip leading spaces */
    while (isspace(*token_s) || *token_s == ';')
	token_s++;

    if (!*token_s) {
	tok.s = NULL;
	return tok;
    }

    tok.s = token_s;
    tok.arg = NULL;
    tok.arglen = 0;

    if (isdigit(*token_s) || *token_s == '.' ||
	((*token_s == '+' || *token_s == '-') && isdigit(*(token_s+1)))) {
	/* It's a number */
	if ( (*token_s == '0' && (*(token_s+1)=='x' || *(token_s+1)=='X'))
				|| strpbrk(token_s, ".eE")==NULL) {
	    *d = (double) strtol(tok.s, &token_s, 0);
	} else {
	    *d = strtod(tok.s, &token_s);
	}
	    
	if (token_s == tok.s || *token_s == ',') {
	    /* Not a valid number */
	    tok.len = 0;
	    return tok;
	}
	tok.len = tok.s - token_s;
	return tok;

    } else if ( (
		(*token_s == 'D') || (*token_s == 'E') ||
		(*token_s == 'S') || (*token_s == '$') ||
		(*token_s == '@')
		) && *(token_s+1)) {
	/* It's define macro / expand macro /
	 * store value / push value / source file.
	 */
	do {
	    ++token_s;
	} while (*token_s && !isspace(*token_s) && *token_s != ';');
	tok.len = token_s - tok.s;
	return tok;

    } else if ((*token_s == '<' || *token_s == '>' || *token_s == '*')
			&& *(token_s+1) == *token_s) {
	/* It's <<, >>, or ** */
	tok.len = 2;
	token_s += 2;
	return tok;

    } else if (isalpha(*token_s) || *token_s == '_') {
	/* It's a name */
	do {
	    ++token_s;
	} while (isalpha(*token_s) || isdigit(*token_s) || *token_s == '_');
	tok.len = token_s - tok.s;
	if (*token_s == ',') {
	    /* There is an argument */
	    tok.arg = ++token_s;
	    while (isalpha(*token_s) || isdigit(*token_s) || *token_s == '_') {
		++token_s;
	    }
	    tok.arglen = token_s - tok.arg;
	}
	return tok;
    } else if (*token_s == '#') {
	/* Rest of line is a comment */
	tok.len = 1;
	token_s += strlen(token_s);
	return tok;
    } else {
	/* It's a single-character op */
	tok.len = 1;
	token_s++;
	return tok;
    }
}

/*
 * A function for applying rpn_expr() to each line of a file.
 * The lines are read using rpn_fgets(), which defaults to fgets(),
 * but can be replaced by the caller with another function.
 * NOTES:
 *	(1) The input buffer we offer has a fixed length of 500.
 *	(2) We use the return value from (*rpn_expr)() as a pointer to
 *	the buffer to use.  This means that your (*rpn_expr)()
 *	function can ignore the passed buffer, and allocate
 *	its own, possibly larger, or dynamically allocated buffer.
 */
double
rpn_file(double val,	/* passed to rpn_expr() */
	int flags,	/* passed to rpn_expr().  If RPN_INIT is one of the
			 * flags, it is passed on the first call to rpn_expr,
			 * then deleted from the flags when successive lines
			 * from fp are processed.  If RPN_ECHO is one of the
			 * flags, we do the echoing ourselves (note that
			 * rpn_expr ignores the RPN_ECHO flag, except to pass
			 * it on to subsidiary calls to rpn_file()).
			 * If RPN_NOFILEOK is one of the flags, it is used,
			 * then deleted from the flags when successive lines
			 * are processed (useful if the caller wants silence
			 * if the initial file doesn't exist, but not if
			 * nested files are missing.)
			 */
	char *tag,	/* Referred to in error messages; typically filename.
			 * See source code for specific uses.
			 */
	FILE *fp,	/* Either open file to process, or NULL, meaning
			 * to open and process the file named by tag.
			 */
	int doprompt,	/* Give prompt before reading each line? */
	char *prefix,	/* passed to rpn_expr() */
	char *msg)	/* passed to rpn_expr() */
{
    char *p, buf[1024], *s;
    double x;
    int n, nexpr;
    static void put_prompt(int, int, FILE *);

    if (debug) {
	fprintf(stderr, "Entering rpn_file\n"
		"\tval = %g\n"     "\tflags = %#o\n"    "\ttag = <%s>\n"
		"\tfile = %p\n"    "\tdoprompt = %d\n"  "\tprefix = <%s>\n"
		"\tmsg = <%s>\n",
		val, flags, tag, fp, doprompt, prefix, msg);
    }

    if (fp == NULL) {
	if (*tag == '~' || strchr(tag, '$')) {
	    /* Use shell to expand filename */
	    s = getenv("SHELL");
	    sprintf(buf, "%s -c 'echo %s'", s ? s : "/bin/sh", tag);
	    if ((fp = popen(buf, "r")) && fgets(buf, sizeof(buf), fp)) {
		n = strlen(buf);
		if (buf[n-1] == '\n')
		    buf[n-1] = '\0';
	    } else {
		if (flags & RPN_NOFILEOK) {
		    *msg = '\0';
		} else {
		    sprintf(msg,
			"%s:rpn_file(): failed to open <%s> for processing: ",
			prefix, tag);
		}
		return val;
	    }
	    if (fp)
		close(fp);
	} else {
	    strcpy(buf, tag);
	}
	if ((fp = fopen(buf, "r")) == NULL) {
	    if (flags & RPN_NOFILEOK) {
		*msg = '\0';
	    } else {
		sprintf(msg,
		    "%s:rpn_file(): failed to open <%s> for processing: ",
		    prefix, buf);
	    }
	    return val;
	}
    }

    flags &= ~RPN_NOFILEOK;

    for (nexpr=1, put_prompt(doprompt, nexpr, stdout);
		(s=rpn_fgets(buf, sizeof(buf), fp));
			nexpr++, put_prompt(doprompt, nexpr, stdout)) {

	if (debug || flags & RPN_ECHO) {
	    /* Be a bit crafty: make sure we print 1 and only 1 newline,
	     * since user-supplied functions replacing rpn_fgets may
	     * not keep the newline.
	     */
	    printf("%s%s [%d]: %s%s", debug ? "(Debug) " : "",
		    tag, nexpr, s, s[strlen(s)-1] == '\n' ? "" : "\n");
	}

	x = rpn_expr(val, flags, &s, 1, prefix, msg);
	if (*msg) {
	    printf("Error on line %d from %s:\n\t", nexpr, tag);
	    puts(msg);
	    *msg = '\0';
	} else {
	    val = x;
	}
	flags &= ~RPN_INIT;
    }
    return val;
}

static void
put_prompt(int doprompt, int i, FILE *fp)
{
    if (doprompt) {
	fprintf(fp, "%d> ", i);
	fflush(fp);
    }
}

#ifdef TEST

int
main(int argc, char **argv)
{
    char buf[500], msg[500];
    double x, d=0.;
    int nexpr=1;

    for(printf("%d> ", nexpr), fflush(stdout);
			fgets(buf, sizeof(buf), stdin);
				printf("%d> ", nexpr++), fflush(stdout)) {
	x = rpn_expr(d, 0, buf, "main", msg);
	if (*msg) {
	    puts(msg);
	} else {
	    d = x;
	    printf("\t%.12g\n", d);
	}
    }
}
#endif
